from time import time
import threading


class Timer:
    _logger = None

    _target_reached_handler = None

    _target_time = None
    target_reached = None
    _timer_thread = None
    _thread_lock = None
    is_stopped = None
    _is_running = None

    _stop_signal = None

    def __init__(self, timeout_method, stop_signal, logger):
        self._target_reached_handler = timeout_method
        self._logger = logger

        self._stop_signal = stop_signal

        self.is_stopped = threading.Event()
        self._is_running = threading.Event()
        self._logger.debug("Timer._timer - is_stopped.set")
        self.is_stopped.set()
        self._logger.debug("Timer._timer - _is_running.clear")
        self._is_running.clear()

        self._timer_thread = None
        self.target_reached = False
        self._thread_lock = threading.Lock()

        self._logger.info("Timer.__init__ - created timer")

    def _get_time_diff(self):
        now = time()
        diff = max(0, self._target_time - now)
        self._logger.debug("Timer._get_time_diff - target time: '{}', now: '{}', diff: '{}'".
                           format(self._target_time, now, diff))
        return diff

    def _timer(self):
        self._logger.debug("Timer._timer - is_stopped.clear")
        self.is_stopped.clear()
        self._logger.debug("Timer._timer - _is_running.set")
        self._is_running.set()

        diff = self._get_time_diff()
        while not self.target_reached and not self._stop_signal.is_set():
            self._logger.info("Timer._timer - waiting '{}' seconds.".format(diff))
            self._stop_signal.wait(diff)
            diff = self._get_time_diff()
            if diff == 0:
                self._logger.info("Timer._timer - timeout")
                self.target_reached = True
                self._target_reached_handler()
            elif not self._stop_signal.is_set():
                self._logger.info("Timer._timer - new targettime applied. waiting {} seconds.".format(diff))
            else:
                self._logger.info("Timer._timer - stop signal")

        self._logger.debug("Timer._timer - _is_running.clear")
        self._is_running.clear()
        self._logger.debug("Timer._timer - is_stopped.set")
        self.is_stopped.set()

    def _start_timer(self):
        self._logger.info("Timer._start_timer - acquiring lock")
        with self._thread_lock:
            if self.is_stopped.is_set():
                self._logger.info("Timer._start_timer - starting timer")
                self.target_reached = False
                self._timer_thread = threading.Thread(target=self._timer)
                self._timer_thread.start()
                self._is_running.wait()
            else:
                self._logger.info("Timer._start_timer - timer already running")

    def set_timer(self, timer_value):
        self._target_time = time() + timer_value
        self._logger.info("Timer.set_timer - waiting '{}' -> set target time to:'{}'.".
                          format(timer_value, self._target_time))
        self._start_timer()



def get_schema():
    return {
            "timerservice": {
                "description": "Root node for thyestes specific entries.",
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "name": {
                            "description": "unique name for timer event",
                            "type": "string"
                        },
                        "topic-sub": {
                            "description": "topics to be subscribed to",
                            "type": "string"
                        },
                        "topic-command": {
                            "description": "message that initiates the timer. optional - if no-entry or None as value "
                                           "every message will trigger timer",
                            "type": "string"
                        },
                        "topic-pub": {
                            "description": "topic to publish the timer event to",
                            "type": "string"
                        },
                        "timer-message": {
                            "description": "message to be published when a timer event happens",
                            "type": "string"
                        },
                        "on-new-command-behavior": {
                            "description": "[restart, spawn] - should the same timer be restarted (there is only one "
                                           "timer and it will be resetted) or should a new timer be spawned (the "
                                           "previous timer continous to count down untouched).",
                            "type": "string",
                            "enum": ["restart", "spawn", "RESTART", "SPAWN", "Restart", "Spawn"]
                        },
                        "timer-value": {
                            "description": "timeout in seconds",
                            "type": "number",
                            "minimum": 0,
                            "exclusiveMinimum": True
                        },
                        "active": {
                            "description": "entry ignored if set to False",
                            "type": "boolean"
                        }
                    },
                    "required": ["active", "name",  "timer-value", "on-new-command-behavior", "timer-message",
                                 "topic-pub", "topic-command", "topic-sub"],
                    "additionalProperties": False
                },
                "additionalProperties": False
            }
           }



import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops.myconfigtools import read_config
from pelops.mylogger import create_logger
import threading
import time
from thyestes.timer import Timer


class TestTimer(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.filename = os.path.abspath(os.path.join(os.path.dirname(__file__), '..')) + "/tests_unit/config.yaml"
        cls.main_config = read_config(cls.filename)
        cls.logger = create_logger(cls.main_config["logger"], "TestTimer")
        cls.logger.info("start ==============================================")

    @classmethod
    def tearDownClass(cls):
        cls.logger.info("end ================================================")

    def setUp(self):
        self.logger.info("----------------------------------------------------")

    def tearDown(self):
        pass

    def test_0init(self):
        self.logger.info("")
        def temp():
            self.assertTrue(False)
        stop_signal = threading.Event()
        stop_signal.clear()
        timer = Timer(temp, stop_signal, self.logger)
        self.assertIsNotNone(timer)
        self.assertTrue(timer.is_stopped.is_set())

    def test_1set_timer(self):
        self.logger.info("test_1set_timer")
        global counter
        counter = 0
        event = threading.Event()
        event.clear()
        stop_signal = threading.Event()
        stop_signal.clear()
        seconds = 1

        def timer_event():
            global counter
            counter += 1
            event.set()

        timer = Timer(timer_event, stop_signal, self.logger)
        self.assertIsNotNone(timer)
        self.assertEqual(counter, 0)
        start = time.time()
        timer.set_timer(seconds)
        self.assertEqual(counter, 0)
        self.assertFalse(timer.is_stopped.is_set())
        time.sleep(seconds/2)
        self.assertEqual(counter, 0)
        self.assertFalse(timer.is_stopped.is_set())
        event.wait(seconds)
        self.assertTrue(event.is_set())
        timer.is_stopped.wait(seconds)
        self.assertTrue(timer.is_stopped.is_set())
        self.assertEqual(counter, 1)
        diff = time.time()-start
        self.assertGreaterEqual(diff, seconds)
        self.assertLess(diff, seconds+1)

    def test_2set_timer_three_times_sequentially(self):
        self.logger.info("test_2set_timer_three_times_sequentially")
        global counter
        counter = 0
        event = threading.Event()
        event.clear()
        stop_signal = threading.Event()
        stop_signal.clear()
        seconds = 0.5

        def timer_event():
            global counter
            counter += 1
            event.set()

        def test(test_timer, c):
            global counter
            self.assertEqual(counter, c-1)
            start = time.time()
            test_timer.set_timer(seconds)
            self.assertEqual(counter, c-1)
            self.assertFalse(test_timer.is_stopped.is_set())
            time.sleep(seconds/2)
            self.assertEqual(counter, c-1)
            self.assertFalse(test_timer.is_stopped.is_set())
            event.wait(seconds)
            self.assertTrue(event.is_set())
            event.clear()
            test_timer.is_stopped.wait(seconds)
            self.assertTrue(test_timer.is_stopped.is_set())
            self.assertEqual(counter, c)
            diff = time.time()-start
            self.assertGreaterEqual(diff, seconds)
            self.assertLess(diff, seconds + 1)

        timer = Timer(timer_event, stop_signal, self.logger)
        self.assertIsNotNone(timer)
        test(timer, 1)
        test(timer, 2)
        test(timer, 3)

    def test_3set_timer_three_times_overlapping(self):
        self.logger.info("test_3set_timer_three_times_overlapping")
        global counter
        counter = 0
        event = threading.Event()
        event.clear()
        stop_signal = threading.Event()
        stop_signal.clear()
        seconds = 1.5
        overlapwait = seconds/3

        def timer_event():
            global counter
            counter += 1
            event.set()

        timer = Timer(timer_event, stop_signal, self.logger)
        self.assertIsNotNone(timer)
        self.assertEqual(counter, 0)
        start = time.time()
        for i in range(3):
            timer.set_timer(seconds)
            self.assertEqual(counter, 0)
            self.assertFalse(timer.is_stopped.is_set())
            time.sleep(overlapwait)

        event.wait(seconds)
        self.assertTrue(event.is_set())
        event.clear()
        timer.is_stopped.wait(seconds)
        self.assertTrue(timer.is_stopped.is_set())
        self.assertEqual(counter, 1)
        diff = time.time() - start
        self.assertGreaterEqual(diff, seconds+overlapwait*2)
        self.assertLess(diff, seconds+overlapwait*2+1)

    def test_4stop_signal(self):
        self.logger.info("test_4stop_signal")
        global counter
        counter = 0
        event = threading.Event()
        event.clear()
        stop_signal = threading.Event()
        stop_signal.clear()
        seconds = 1

        def timer_event():
            global counter
            counter += 1
            event.set()

        timer = Timer(timer_event, stop_signal, self.logger)
        self.assertIsNotNone(timer)
        self.assertEqual(counter, 0)
        start = time.time()
        timer.set_timer(seconds)
        self.assertEqual(counter, 0)
        self.assertFalse(timer.is_stopped.is_set())
        time.sleep(seconds/2)
        self.assertEqual(counter, 0)
        self.assertFalse(timer.is_stopped.is_set())
        stop_signal.set()
        timer.is_stopped.wait(seconds)
        self.assertTrue(timer.is_stopped.is_set())
        self.assertEqual(counter, 0)
        event.wait(seconds)
        self.assertFalse(event.is_set())


if __name__ == '__main__':
    unittest.main()



from pelops.abstractmicroservice import AbstractMicroservice
import thyestes
import thyestes.schema.schema
import thyestes.timerfactory


class Timerservice(AbstractMicroservice):
    _version = thyestes.version
    _controller = None

    def __init__(self, config, mqtt_client=None, logger=None):
        """
        Constructor - creates the services and the tasks

        :param config: config yaml structure
        :param mqtt_client: mqtt client instance
        :param logger: logger instance
        """
        AbstractMicroservice.__init__(self, config, "timerservice", mqtt_client, logger)
        self._controller = thyestes.timerfactory.TimerFactory.create_controller_list(self._config, self._mqtt_client,
                                                                                     self._logger)

    def _start(self):
        for controller in self._controller:
            controller.start()

    def _stop(self):
        for controller in self._controller:
            controller.stop()

    @classmethod
    def _get_description(cls):
        return "Thyestes it a timer microservice. Listens on topics for specific messages, starts a timer when such " \
               "a messages has been received and publishes a predefined message after the timer expired."

    @classmethod
    def _get_schema(cls):
        """
        Get the sub schema to validate the yaml-config file against.

        :return: json-schema dict
        """
        return thyestes.schema.schema.get_schema()


def standalone():
    Timerservice.standalone()


if __name__ == "__main__":
    Timerservice.standalone()



import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops.mymqttclient import MyMQTTClient
from pelops.myconfigtools import read_config
from pelops.mylogger import create_logger
from thyestes.timerfactory import TimerFactory
from thyestes.spawntimer import SpawnTimer
from thyestes.restarttimer import RestartTimer


class TestTimerFactory(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.filename = os.path.abspath(os.path.join(os.path.dirname(__file__), '..')) + "/tests_unit/config.yaml"
        cls.main_config = read_config(cls.filename)
        cls.logger = create_logger(cls.main_config["logger"], "TestTimerFactory")
        cls.mqttclient = MyMQTTClient(cls.main_config["mqtt"], cls.logger, True)
        cls.logger.info("start ==============================================")

    @classmethod
    def tearDownClass(cls):
        cls.logger.info("end ================================================")

    def setUp(self):
        self.logger.info("----------------------------------------------------")

    def tearDown(self):
        pass

    def test_0create_controller_restarttimer(self):
        self.logger.info("test_0create_controller_restarttimer")
        c = TimerFactory.create_controller(self.main_config["timerservice"][0], self.mqttclient, self.logger)
        self.assertIsNotNone(c)
        self.assertTrue(type(c), RestartTimer)

    def test_1create_controller_spawntimer(self):
        self.logger.info("test_1create_controller_spawntimer")
        c = TimerFactory.create_controller(self.main_config["timerservice"][1], self.mqttclient, self.logger)
        self.assertIsNotNone(c)
        self.assertTrue(type(c), SpawnTimer)

    def test_2create_controller_inactive(self):
        self.logger.info("test_2create_controller_inactive")
        c = TimerFactory.create_controller(self.main_config["timerservice"][2], self.mqttclient, self.logger)
        self.assertIsNone(c)

    def test_3create_controller_unknown(self):
        self.logger.info("test_3create_controller_unknown")
        conf = self.main_config["timerservice"][1].copy()
        conf["on-new-command-behavior"] = "somerandomtext"
        with self.assertRaises(ValueError):
            c = TimerFactory.create_controller(conf, self.mqttclient, self.logger)

    def test_4create_controllers(self):
        self.logger.info("test_4create_controllers")
        c = TimerFactory.create_controller_list(self.main_config["timerservice"], self.mqttclient, self.logger)
        self.assertIsNotNone(c)
        self.assertEqual(len(c), 2)


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops.mymqttclient import MyMQTTClient
from pelops.myconfigtools import read_config
from pelops.mylogger import create_logger
import threading
import time
from thyestes.restarttimer import RestartTimer


class TestRestartTimer(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.filename = os.path.abspath(os.path.join(os.path.dirname(__file__), '..')) + "/tests_unit/config.yaml"
        cls.main_config = read_config(cls.filename)
        cls.logger = create_logger(cls.main_config["logger"], "TestRestartTimer")
        cls.mqttclient = MyMQTTClient(cls.main_config["mqtt"], cls.logger, True)
        cls.mqttclient.connect()

        cls.logger.info("start ==============================================")

    @classmethod
    def tearDownClass(cls):
        cls.logger.info("end ================================================")
        cls.mqttclient.disconnect()

    def setUp(self):
        self.logger.info("----------------------------------------------------")
        self.config = {
            "name": "test",
            "topic-sub": "/test/1/event",
            "topic-command": "start",
            "topic-pub": "/test/1/timer",
            "timer-message": "timer",
            "timer-value": 1
        }

    def tearDown(self):
        self.mqttclient.unsubscribe_all()

    def test_0init(self):
        self.logger.info("test_0init")
        rt = RestartTimer(self.config, self.mqttclient, self.logger)
        self.assertIsNotNone(rt)
        self.assertIsNotNone(rt._timer)
        self.assertEqual(rt._timer_value, float(self.config["timer-value"]))
        self.assertEqual(rt._topic_sub, self.config["topic-sub"])
        self.assertEqual(rt._topic_command, self.config["topic-command"])
        self.assertEqual(rt._topic_pub, self.config["topic-pub"])
        self.assertEqual(rt._timer_message, self.config["timer-message"])

    def test_1start_stop(self):
        self.logger.info("test_1start_stop")
        rt = RestartTimer(self.config, self.mqttclient, self.logger)
        self.assertIsNotNone(rt)
        self.assertTrue(rt._stop_timer.is_set())
        self.assertTrue(rt._timer.is_stopped.is_set())
        rt.start()
        self.assertFalse(rt._stop_timer.is_set())
        self.assertTrue(rt._timer.is_stopped.is_set())
        rt.stop()
        self.assertTrue(rt._stop_timer.is_set())
        self.assertTrue(rt._timer.is_stopped.is_set())

    def test_2set_timer_once_happy(self):
        self.logger.info("test_2set_timer_once_happy")
        global counter
        counter = 0
        event = threading.Event()
        event.clear()

        def mqtt_handler(value):
            value = value.decode()
            self.assertEqual(value, self.config["timer-message"])
            global counter
            counter += 1
            event.set()

        self.mqttclient.subscribe(self.config["topic-pub"], mqtt_handler)
        rt = RestartTimer(self.config, self.mqttclient, self.logger)
        rt.start()
        self.assertEqual(counter, 0)
        self.mqttclient.publish(self.config["topic-sub"], self.config["topic-command"])
        rt._timer._is_running.wait(1)
        self.assertTrue(rt._timer._is_running.is_set())
        event.wait(5)
        self.assertTrue(event.is_set())
        self.assertEqual(counter, 1)
        event.clear()
        rt.stop()
        self.assertFalse(rt._timer._is_running.is_set())
        self.assertTrue(rt._timer.is_stopped.is_set())
        self.mqttclient.unsubscribe(self.config["topic-pub"], mqtt_handler)
        self.assertEqual(counter, 1)
        self.assertFalse(event.is_set())

    def test_3set_timer_once_onebad(self):
        self.logger.info("test_3set_timer_once_onebad")
        global counter
        counter = 0
        event = threading.Event()
        event.clear()

        def mqtt_handler(value):
            value = value.decode()
            self.assertEqual(value, self.config["timer-message"])
            global counter
            counter += 1
            event.set()

        self.mqttclient.subscribe(self.config["topic-pub"], mqtt_handler)
        rt = RestartTimer(self.config, self.mqttclient, self.logger)
        rt.start()
        self.assertEqual(counter, 0)
        self.mqttclient.publish(self.config["topic-sub"], self.config["topic-command"]+"somerandomtext")
        rt._timer._is_running.wait(1)
        self.assertFalse(rt._timer._is_running.is_set())
        event.wait(1)
        self.assertFalse(event.is_set())
        self.assertEqual(counter, 0)
        rt.stop()
        self.mqttclient.unsubscribe(self.config["topic-pub"], mqtt_handler)
        self.assertEqual(counter, 0)
        self.assertFalse(event.is_set())

    def test_4set_timer_once_nofilter(self):
        self.logger.info("test_4set_timer_once_nofilter")
        global counter
        counter = 0
        event = threading.Event()
        event.clear()

        def mqtt_handler(value):
            value = value.decode()
            self.assertEqual(value, self.config["timer-message"])
            global counter
            counter += 1
            event.set()

        self.config["topic-command"] = None
        self.mqttclient.subscribe(self.config["topic-pub"], mqtt_handler)
        rt = RestartTimer(self.config, self.mqttclient, self.logger)
        self.assertIsNone(rt._topic_command)
        rt.start()
        self.assertEqual(counter, 0)
        self.mqttclient.publish(self.config["topic-sub"], "somerandomtext")
        rt._timer._is_running.wait(1)
        self.assertTrue(rt._timer._is_running.is_set())
        event.wait(5)
        self.assertTrue(event.is_set())
        self.assertEqual(counter, 1)
        event.clear()
        rt.stop()
        self.assertFalse(rt._timer._is_running.is_set())
        self.assertTrue(rt._timer.is_stopped.is_set())
        self.mqttclient.unsubscribe(self.config["topic-pub"], mqtt_handler)
        self.assertEqual(counter, 1)
        self.assertFalse(event.is_set())

    def test_5set_timer_three_times_sequentially(self):
        self.logger.info("test_5set_timer_three_times_sequentially")
        global counter
        counter = 0
        event = threading.Event()
        event.clear()

        def mqtt_handler(value):
            value = value.decode()
            self.assertEqual(value, self.config["timer-message"])
            global counter
            counter += 1
            event.set()

        self.mqttclient.subscribe(self.config["topic-pub"], mqtt_handler)
        rt = RestartTimer(self.config, self.mqttclient, self.logger)
        rt.start()

        for i in range(3):
            self.assertEqual(counter, i)
            self.mqttclient.publish(self.config["topic-sub"], self.config["topic-command"])
            rt._timer._is_running.wait(1)
            self.assertTrue(rt._timer._is_running.is_set())
            event.wait(5)
            self.assertTrue(event.is_set())
            self.assertEqual(counter, i+1)
            event.clear()

        rt.stop()
        self.assertFalse(rt._timer._is_running.is_set())
        self.assertTrue(rt._timer.is_stopped.is_set())
        self.mqttclient.unsubscribe(self.config["topic-pub"], mqtt_handler)
        self.assertEqual(counter, 3)
        self.assertFalse(event.is_set())

    def test_6set_timer_three_times_overlapping(self):
        self.logger.info("test_6set_timer_three_times_overlapping")
        global counter
        counter = 0
        event = threading.Event()
        event.clear()

        def mqtt_handler(value):
            value = value.decode()
            self.assertEqual(value, self.config["timer-message"])
            global counter
            counter += 1
            event.set()

        self.config["timer-value"] = 1
        self.mqttclient.subscribe(self.config["topic-pub"], mqtt_handler)
        rt = RestartTimer(self.config, self.mqttclient, self.logger)
        rt.start()
        start = time.time()

        for i in range(3):
            self.mqttclient.publish(self.config["topic-sub"], self.config["topic-command"])
            time.sleep(0.5)
            self.assertFalse(event.is_set())
            self.assertEqual(counter, 0)

        event.wait(3)
        self.assertTrue(event.is_set())
        event.clear()
        diff = time.time() - start
        self.assertEqual(counter, 1)
        rt._timer.is_stopped.wait(1)
        self.assertTrue(rt._timer.is_stopped.is_set())
        self.assertGreaterEqual(diff, 2)
        self.assertLess(diff, 3)

        rt.stop()
        self.assertFalse(rt._timer._is_running.is_set())
        self.assertTrue(rt._timer.is_stopped.is_set())
        self.mqttclient.unsubscribe(self.config["topic-pub"], mqtt_handler)
        self.assertEqual(counter, 1)
        self.assertFalse(event.is_set())

    def test_7set_timer_stop_signal(self):
        self.logger.info("test_7set_timer_stop_signal")
        global counter
        counter = 0
        event = threading.Event()
        event.clear()

        def mqtt_handler(value):
            value = value.decode()
            self.assertEqual(value, self.config["timer-message"])
            global counter
            counter += 1
            event.set()

        self.config["timer-value"] = 1
        self.mqttclient.subscribe(self.config["topic-pub"], mqtt_handler)
        rt = RestartTimer(self.config, self.mqttclient, self.logger)
        rt.start()
        self.mqttclient.publish(self.config["topic-sub"], self.config["topic-command"])

        time.sleep(0.5)
        self.assertFalse(event.is_set())
        self.assertEqual(counter, 0)

        rt.stop()
        self.assertFalse(rt._timer._is_running.is_set())
        self.assertTrue(rt._timer.is_stopped.is_set())
        self.mqttclient.unsubscribe(self.config["topic-pub"], mqtt_handler)
        self.assertEqual(counter, 0)
        event.wait(1)
        self.assertFalse(event.is_set())


if __name__ == '__main__':
    unittest.main()



import pelops.mylogger
import thyestes.restarttimer
import thyestes.spawntimer


"""


"""


class TimerFactory:
    """
    Factory class - creates silblings from AElement based on the provided config yaml structure.

    config yaml - everything from acontroller and:
        on-new-command-behavior: restart  # [restart, spawn] - should the same timer be restarted (there is only one
                                            timer and it will be resetted) or should a new timer be spawned (the
                                            previous timer continous to count down untouched).
        active: True  # entry ignored if set to False
    """

    @staticmethod
    def create_controller(config, mqtt_client, logger):
        """
        Create the element that corresponds to the provided config yaml.

        :param config: config yaml structure for a single element
        :param mqtt_client: mymqttclient instance
        :param logger: logger instance
        :return: instance of the created element
        """
        _logger = pelops.mylogger.get_child(logger, __name__)
        controller = None
        if config["active"]:
            if config["on-new-command-behavior"].lower() == "spawn":
                controller = thyestes.spawntimer.SpawnTimer(config, mqtt_client, logger)
            elif config["on-new-command-behavior"].lower() == "restart":
                controller = thyestes.restarttimer.RestartTimer(config, mqtt_client, logger)
            else:
                _logger.error("TimerFactory.create_controller - unknown behavior '{}'".
                              format(config["on-new-command-behavior"].lower()))
                raise ValueError("TimerFactory.create_controller - unknown behavior '{}'".
                                 format(config["on-new-command-behavior"].lower()))
        else:
            _logger.info("TimerFactory.create_controller - skipping inactive element '{}.{}'.".
                      format(config["on-new-command-behavior"].lower(), config["name"]))

        return controller

    @staticmethod
    def create_controller_list(configs, mqtt_client, logger):
        """
        Create all controller that are defined in the provided config.

        :param configs: config yaml for timer (array)
        :param mqtt_client: mymqttclient instance
        :param logger: logger instance
        :return: list of all active elements
        """
        controller_list = []
        _logger = pelops.mylogger.get_child(logger, __name__)

        _logger.info("TimerFactory.create_controller - start")

        for config in configs:
            controller = TimerFactory.create_controller(config, mqtt_client, logger)
            if controller is not None:
                controller_list.append(controller)

        _logger.info("TimerFactory.create_controller - created {} controller.".format(len(controller_list)))
        _logger.info("TimerFactory.create_controller - finished")

        return controller_list




from thyestes.acontroller import AController
from thyestes.timer import Timer
from collections import deque


class SpawnTimer(AController):
    _timer_list = None

    def __init__(self, config, mqtt_client, logger):
        AController.__init__(self, config, mqtt_client, logger)

        self._timer_list = deque()
        self._logger.info("SpawnTimer.__init__ - finished constructor")

    def _update_timer(self):
        self._logger.info("SpawnTimer._update_timer - adding new timer to deque")
        timer = Timer(self._timer_event, self._stop_timer, self._logger)
        timer.set_timer(self._timer_value)
        self._timer_list.append(timer)
        self._logger.info("SpawnTimer._update_timer - active timer: {}.".format(len(self._timer_list)))

    def _timer_event(self):
        self._logger.info("SpawnTimer._timer_event")
        self._purge_deque()
        self._publish_timer_message()

    def _purge_deque(self):
        """
        purge the deque from all stopped timers.

        notable behavior: the current timer, that fired _timer_event has reached its timeout but has not been stopped
        yet. in fact, i can only be closed after this method has been processed. thus, it will be in this list until the
        next timer fires and _purge_deque is called again. concluding, after the first timer has been created, this
        deque will always have at least on entry.
        """
        self._logger.info("SpawnTimer._purge_deque - removing all stopped timer")
        i = 0
        while len(self._timer_list) > 0 and self._timer_list[0].is_stopped.is_set():
            self._logger.debug("SpawnTimer._purge_deque - removing timer")
            self._timer_list.popleft()
            i+=1
        self._logger.info("SpawnTimer._purge_deque - removed {} expired timer from list (remaining timer: {})".
                          format(i, len(self._timer_list)))

    def _start(self):
        pass

    def _stop(self):
        while len(self._timer_list) > 0:
            self._timer_list[0].is_stopped.wait()
            self._purge_deque()









from thyestes.acontroller import AController
from thyestes.timer import Timer


class RestartTimer(AController):
    _timer = None

    def __init__(self, config, mqtt_client, logger):
        AController.__init__(self, config, mqtt_client, logger)

        self._timer = Timer(self._publish_timer_message, self._stop_timer, self._logger)
        self._logger.info("RestartTimer.__init__ - finished constructor")

    def _update_timer(self):
        self._timer.set_timer(self._timer_value)

    def _start(self):
        pass

    def _stop(self):
        self._timer.is_stopped.wait()







import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops.mymqttclient import MyMQTTClient
from pelops.myconfigtools import read_config
from pelops.mylogger import create_logger
from thyestes.timerservice import Timerservice
from thyestes.spawntimer import SpawnTimer
from thyestes.restarttimer import RestartTimer
import threading
import time


class TestTimerService(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.filename = os.path.abspath(os.path.join(os.path.dirname(__file__), '..')) + "/tests_unit/config.yaml"
        cls.main_config = read_config(cls.filename)
        cls.logger = create_logger(cls.main_config["logger"], "TestTimerService")
        cls.mqttclient = MyMQTTClient(cls.main_config["mqtt"], cls.logger, True)
        cls.logger.info("start ==============================================")

    @classmethod
    def tearDownClass(cls):
        cls.logger.info("end ================================================")

    def setUp(self):
        self.logger.info("----------------------------------------------------")

    def tearDown(self):
        pass

    def test_0init(self):
        self.logger.info("test_0init")
        ts = Timerservice(self.main_config, self.mqttclient, self.logger)
        self.assertIsNotNone(ts)
        self.assertEqual(len(ts._controller), 2)
        self.assertEqual(type(ts._controller[0]), RestartTimer)
        self.assertEqual(type(ts._controller[1]), SpawnTimer)

    def test_1startstop(self):
        self.logger.info("test_1startstop")
        ts = Timerservice(self.main_config, self.mqttclient, self.logger)

        self.assertTrue(ts._controller[0]._stop_timer.is_set())
        self.assertTrue(ts._controller[0]._timer.is_stopped.is_set())
        self.assertTrue(ts._controller[1]._stop_timer.is_set())
        self.assertEqual(len(ts._controller[1]._timer_list), 0)
        ts.start()
        self.assertFalse(ts._controller[0]._stop_timer.is_set())
        self.assertTrue(ts._controller[0]._timer.is_stopped.is_set())
        self.assertEqual(len(ts._controller[1]._timer_list), 0)
        ts.stop()
        self.assertTrue(ts._controller[0]._stop_timer.is_set())
        self.assertTrue(ts._controller[0]._timer.is_stopped.is_set())
        self.assertTrue(ts._controller[1]._stop_timer.is_set())
        self.assertEqual(len(ts._controller[1]._timer_list), 0)

    def test_2activate_restarttimer(self):
        self.logger.info("test_2activate_restarttimer")
        global counter_spawn
        counter_spawn = 0
        event_spawn = threading.Event()
        event_spawn.clear()

        global counter_restart
        counter_restart = 0
        event_restart = threading.Event()
        event_restart.clear()

        def mqtt_handler_spawn(value):
            value = value.decode()
            self.assertEqual(value, self.main_config["timerservice"][1]["timer-message"])
            global counter_spawn
            counter_spawn += 1
            event_spawn.set()

        def mqtt_handler_restart(value):
            value = value.decode()
            self.assertEqual(value, self.main_config["timerservice"][0]["timer-message"])
            global counter_restart
            counter_restart += 1
            event_restart.set()

        self.mqttclient.subscribe(self.main_config["timerservice"][0]["topic-pub"], mqtt_handler_restart)
        self.mqttclient.subscribe(self.main_config["timerservice"][1]["topic-pub"], mqtt_handler_spawn)

        ts = Timerservice(self.main_config, self.mqttclient, self.logger)
        ts.start()
        self.assertFalse(event_restart.is_set())
        self.assertFalse(event_spawn.is_set())
        self.assertEqual(counter_restart, 0)
        self.assertEqual(counter_spawn, 0)
        self.mqttclient.publish(self.main_config["timerservice"][0]["topic-sub"],
                                self.main_config["timerservice"][0]["topic-command"])
        event_spawn.wait(3)
        self.assertFalse(event_spawn.is_set())
        event_restart.wait(3)
        self.assertTrue(event_restart.is_set())
        event_restart.clear()
        self.assertEqual(counter_restart, 1)
        self.assertEqual(counter_spawn, 0)
        ts.stop()

        self.mqttclient.unsubscribe(self.main_config["timerservice"][0]["topic-pub"], mqtt_handler_restart)
        self.mqttclient.unsubscribe(self.main_config["timerservice"][1]["topic-pub"], mqtt_handler_spawn)
        self.assertFalse(event_restart.is_set())
        self.assertFalse(event_spawn.is_set())
        self.assertEqual(counter_restart, 1)
        self.assertEqual(counter_spawn, 0)

    def test_3activate_spawntimer(self):
        self.logger.info("test_3activate_spawntimer")
        global counter_spawn
        counter_spawn = 0
        event_spawn = threading.Event()
        event_spawn.clear()

        global counter_restart
        counter_restart = 0
        event_restart = threading.Event()
        event_restart.clear()

        def mqtt_handler_spawn(value):
            value = value.decode()
            self.assertEqual(value, self.main_config["timerservice"][1]["timer-message"])
            global counter_spawn
            counter_spawn += 1
            event_spawn.set()

        def mqtt_handler_restart(value):
            value = value.decode()
            self.assertEqual(value, self.main_config["timerservice"][0]["timer-message"])
            global counter_restart
            counter_restart += 1
            event_restart.set()

        self.mqttclient.subscribe(self.main_config["timerservice"][0]["topic-pub"], mqtt_handler_restart)
        self.mqttclient.subscribe(self.main_config["timerservice"][1]["topic-pub"], mqtt_handler_spawn)

        ts = Timerservice(self.main_config, self.mqttclient, self.logger)
        ts.start()
        self.assertFalse(event_restart.is_set())
        self.assertFalse(event_spawn.is_set())
        self.assertEqual(counter_restart, 0)
        self.assertEqual(counter_spawn, 0)
        self.mqttclient.publish(self.main_config["timerservice"][1]["topic-sub"],
                                self.main_config["timerservice"][1]["topic-command"])
        event_spawn.wait(3)
        self.assertTrue(event_spawn.is_set())
        event_spawn.clear()
        event_restart.wait(3)
        self.assertFalse(event_restart.is_set())
        self.assertEqual(counter_restart, 0)
        self.assertEqual(counter_spawn, 1)
        ts.stop()

        self.mqttclient.unsubscribe(self.main_config["timerservice"][0]["topic-pub"], mqtt_handler_restart)
        self.mqttclient.unsubscribe(self.main_config["timerservice"][1]["topic-pub"], mqtt_handler_spawn)
        self.assertFalse(event_restart.is_set())
        self.assertFalse(event_spawn.is_set())
        self.assertEqual(counter_restart, 0)
        self.assertEqual(counter_spawn, 1)

    def test_4activate_both(self):
        self.logger.info("test_4activate_both")
        global counter_spawn
        counter_spawn = 0
        event_spawn = threading.Event()
        event_spawn.clear()

        global counter_restart
        counter_restart = 0
        event_restart = threading.Event()
        event_restart.clear()

        def mqtt_handler_spawn(value):
            value = value.decode()
            self.assertEqual(value, self.main_config["timerservice"][1]["timer-message"])
            global counter_spawn
            counter_spawn += 1
            event_spawn.set()

        def mqtt_handler_restart(value):
            value = value.decode()
            self.assertEqual(value, self.main_config["timerservice"][0]["timer-message"])
            global counter_restart
            counter_restart += 1
            event_restart.set()

        self.mqttclient.subscribe(self.main_config["timerservice"][0]["topic-pub"], mqtt_handler_restart)
        self.mqttclient.subscribe(self.main_config["timerservice"][1]["topic-pub"], mqtt_handler_spawn)

        ts = Timerservice(self.main_config, self.mqttclient, self.logger)
        ts.start()
        self.assertFalse(event_restart.is_set())
        self.assertFalse(event_spawn.is_set())
        self.assertEqual(counter_restart, 0)
        self.assertEqual(counter_spawn, 0)
        self.mqttclient.publish(self.main_config["timerservice"][0]["topic-sub"],
                                self.main_config["timerservice"][0]["topic-command"])
        self.mqttclient.publish(self.main_config["timerservice"][1]["topic-sub"],
                                self.main_config["timerservice"][1]["topic-command"])
        event_spawn.wait(3)
        self.assertTrue(event_spawn.is_set())
        event_spawn.clear()
        event_restart.wait(3)
        self.assertTrue(event_restart.is_set())
        event_restart.clear()
        self.assertEqual(counter_restart, 1)
        self.assertEqual(counter_spawn, 1)
        ts.stop()

        self.mqttclient.unsubscribe(self.main_config["timerservice"][0]["topic-pub"], mqtt_handler_restart)
        self.mqttclient.unsubscribe(self.main_config["timerservice"][1]["topic-pub"], mqtt_handler_spawn)
        self.assertFalse(event_restart.is_set())
        self.assertFalse(event_spawn.is_set())
        self.assertEqual(counter_restart, 1)
        self.assertEqual(counter_spawn, 1)

    def test_5stop_both(self):
        self.logger.info("test_5stop_both")
        global counter_spawn
        counter_spawn = 0
        event_spawn = threading.Event()
        event_spawn.clear()

        global counter_restart
        counter_restart = 0
        event_restart = threading.Event()
        event_restart.clear()

        def mqtt_handler_spawn(value):
            value = value.decode()
            self.assertEqual(value, self.main_config["timerservice"][1]["timer-message"])
            global counter_spawn
            counter_spawn += 1
            event_spawn.set()

        def mqtt_handler_restart(value):
            value = value.decode()
            self.assertEqual(value, self.main_config["timerservice"][0]["timer-message"])
            global counter_restart
            counter_restart += 1
            event_restart.set()

        self.mqttclient.subscribe(self.main_config["timerservice"][0]["topic-pub"], mqtt_handler_restart)
        self.mqttclient.subscribe(self.main_config["timerservice"][1]["topic-pub"], mqtt_handler_spawn)

        ts = Timerservice(self.main_config, self.mqttclient, self.logger)
        ts.start()
        self.assertFalse(event_restart.is_set())
        self.assertFalse(event_spawn.is_set())
        self.assertEqual(counter_restart, 0)
        self.assertEqual(counter_spawn, 0)

        time.sleep(0.5)
        self.assertFalse(event_restart.is_set())
        self.assertFalse(event_spawn.is_set())
        self.assertEqual(counter_restart, 0)
        self.assertEqual(counter_spawn, 0)
        ts.stop()

        self.mqttclient.publish(self.main_config["timerservice"][0]["topic-sub"],
                                self.main_config["timerservice"][0]["topic-command"])
        self.mqttclient.publish(self.main_config["timerservice"][1]["topic-sub"],
                                self.main_config["timerservice"][1]["topic-command"])

        event_spawn.wait(3)
        self.assertFalse(event_spawn.is_set())
        event_restart.wait(3)
        self.assertFalse(event_restart.is_set())
        self.assertEqual(counter_restart, 0)
        self.assertEqual(counter_spawn, 0)

        self.mqttclient.unsubscribe(self.main_config["timerservice"][0]["topic-pub"], mqtt_handler_restart)
        self.mqttclient.unsubscribe(self.main_config["timerservice"][1]["topic-pub"], mqtt_handler_spawn)
        self.assertFalse(event_restart.is_set())
        self.assertFalse(event_spawn.is_set())
        self.assertEqual(counter_restart, 0)
        self.assertEqual(counter_spawn, 0)


if __name__ == '__main__':
    unittest.main()




import threading
import pelops.mylogger


class AController:
    """...

    config:
        name: eggtimer  # unique name for timer
        topic-sub: /test/1/event  # topics to be subscribed to
        topic-command: start  # message that initiates the timer. optional - if no-entry or None as value every message will trigger timer
        topic-pub: /test/1/timer  # topic to publish the timer event to
        timer-message: timer  # message to be published when a timer event happens
        timer-value: 30  # timeout in seconds
    """

    _timer_value = -1  # poll time in seconds
    _stop_timer = None  # threading.Event to signal the poll loop to stop immediately.
    _timer_stopped = None

    _config = False
    _logger = False  # print debugging information if set to yes.
    _mqtt_client = False  # mqtt client instance

    def __init__(self, config, mqtt_client, logger):
        self._config = config
        self._logger = pelops.mylogger.get_child(logger, self._config["name"])
        self._mqtt_client = mqtt_client

        self._logger.info("AController.__init__ - starting constructor")
        self._logger.debug("AController.__init__ - config: '{}'".format(self._config))

        self._timer_value = float(self._config["timer-value"])
        self._topic_sub = self._config["topic-sub"]
        try:
            self._topic_command = self._config["topic-command"]
            if not self._topic_command or len(self._topic_command) == 0:
                self._topic_command = None
        except KeyError:
            self._topic_command = None
        self._topic_pub = self._config["topic-pub"]
        self._timer_message = self._config["timer-message"]

        self._stop_timer = threading.Event()
        self._stop_timer.set()

    def _message_handler(self, value):
        value = value.decode()
        self._logger.info("_message_handler - received message")
        self._logger.debug("_message_handler - message content: '{}'".format(value))
        if not self._topic_command or self._topic_command == value:
            self._logger.info("_message_handler - processing message")
            self._update_timer()
        else:
            self._logger.info("_message_handler - skipping message")
            self._logger.debug("_message_handler - content of message is not equivalent to expected value "
                               "('{}')".format(self._topic_command))

    def _update_timer(self):
        raise NotImplementedError

    def start(self):
        self._logger.info("AController.start - starting controller.")
        self._stop_timer.clear()
        self._start()
        self._mqtt_client.subscribe(self._topic_sub, self._message_handler)
        self._logger.info("AController.start - controller started.")

    def _start(self):
        raise NotImplementedError

    def _publish_timer_message(self):
        self._logger.info("AController._publish_timer_message - publishing '{}' to topic '{}'.".
                          format(self._timer_message, self._topic_pub))
        self._mqtt_client.publish(self._topic_pub, self._timer_message)

    def stop(self):
        self._logger.info("AController.stop - stopping controller.")
        self._mqtt_client.unsubscribe(self._topic_sub, self._message_handler)
        self._stop_timer.set()
        self._stop()
        self._logger.info("AController.stop - controller stopped.")

    def _stop(self):
        raise NotImplementedError

